# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
# ================================================================================================

from libcpp.vector cimport vector

cdef extern from "cpu_sa.h":
    vector[double] general_simulated_annealing(
            char*,  # giant array that holds all states
            const int,  # number of samples
            vector[double] &,  # h
            vector[int] &,  # coupler starts
            vector[int] &,  # coupler ends
            vector[double] &,  # coupler weights
            int, # sweeps per beta
            vector[double] &,  # beta schedule
            unsigned long long)  # seed
